#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_plumgrid.neutron.plugins.extensions import \
    physical_attachment_point as ext_pap
from neutron.db import common_db_mixin
from neutron.db import model_base
from neutron.db import models_v2
from oslo_log import log as logging
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import exc

LOG = logging.getLogger(__name__)


class PhysicalAttachmentPoint(model_base.BASEV2, models_v2.HasId,
                              models_v2.HasTenant):
    """DB definition for PLUMgrid physical attachment point object"""

    __tablename__ = "pg_physical_attachment_points"

    name = sa.Column(sa.String(255))
    transit_domain_id = sa.Column(sa.String(255))
    hash_mode = sa.Column(sa.String(255))
    lacp = sa.Column(sa.Boolean)
    active_standby = sa.Column(sa.Boolean, nullable=False, default=False)
    implicit = sa.Column(sa.Boolean)


class Interface(model_base.BASEV2, models_v2.HasId):
    """
    DB definition for interfaces object required by PLUMgrid
    Physical Attachment Points.
    """

    __tablename__ = "pg_physical_attachment_point_interfaces"

    hostname = sa.Column(sa.String(255))
    interface = sa.Column(sa.String(255))
    pap_id = sa.Column(sa.String(36),
                       sa.ForeignKey("pg_physical_attachment_points.id",
                                     ondelete="CASCADE"),
                       nullable=False)
    pap = orm.relationship(PhysicalAttachmentPoint,
              backref=orm.backref("interfaces",
                                  lazy="joined",
                                  cascade="all,delete"),
              primaryjoin="PhysicalAttachmentPoint.id==Interface.pap_id")


class PhysicalAttachmentPointDb(common_db_mixin.CommonDbMixin):
    def create_physical_attachment_point(self, context,
            physical_attachment_point):
        """
        creates a physical attachment point with initial set of interfaces
        Args:
             physical_attachment_point:
                   JSON object with physical attachment point attributes
                   name : display name of physical attachment point
                   tenant_id : tenant uuid
                   id : physical attachment point
                   hash_mode : L2, L3 etc
                   lacp : True/False
                   active_standby: True/False
                   interfaces : List of interfaces to be attached to physical
                                attachment point
        """
        pap = physical_attachment_point["physical_attachment_point"]

        #  check availability of physical interfaces
        self._check_ifc(context, pap["interfaces"])

        #  check if transit domain already contains
        #  a physical attachment point
        tvd_id = pap["transit_domain_id"]
        t_limit = self._check_transit_domain_limit(context, tvd_id)

        if not t_limit:
            if pap.get("implicit") is None:
                pap["implicit"] = False

            lacp_flag = self._lacp_flag(pap)
            if pap["hash_mode"].lower() != "l2":
                if not lacp_flag:
                    raise ext_pap.InvalidLacpValue(hash_mode=pap["hash_mode"])

            as_flag = self._as_flag(pap)
            if as_flag and lacp_flag:
                raise ext_pap.InvalidLacpValueAs

            with context.session.begin(subtransactions=True):
                pap_db = PhysicalAttachmentPoint(
                             tenant_id=pap["tenant_id"],
                             name=pap["name"],
                             hash_mode=pap["hash_mode"],
                             lacp=lacp_flag,
                             active_standby=as_flag,
                             transit_domain_id=pap["transit_domain_id"],
                             implicit=pap["implicit"])

                context.session.add(pap_db)
                for interface in pap["interfaces"]:
                    interface_db = Interface(hostname=interface["hostname"],
                                             interface=interface["interface"],
                                             pap=pap_db)

                    context.session.add(interface_db)
            return self._make_pap_dict(pap_db)

    def update_physical_attachment_point(self, context, pap_id,
            physical_attachment_point):
        """
        Updates an existing physical attachment point
        Args:
             pap_id:
                   uuid of the physical attachment point being updated
             physical_attachment_point:
                   JSON with updated attributes of the physical
                   attachment point
                   name : display name of physical attachment point
                   tenant_id : tenant uuid
                   id : physical attachment point
                   hash_mode : L2+L3 etc
                   lacp : True/False
                   active_standby: True/False
                   interfaces : List of interfaces to be attached to physical
                                attachment point
        """

        pap = physical_attachment_point["physical_attachment_point"]
        if not pap:
            raise ext_pap.UpdateParametersRequired()
        with context.session.begin(subtransactions=True):
            try:
                query = self._model_query(context, PhysicalAttachmentPoint)
                pap_db = query.filter_by(id=pap_id).one()
            except exc.NoResultFound:
                raise ext_pap.NoPhysicalAttachmentPointFound(id=pap_id)

            if 'name' in pap:
                pap_db.update({"name": pap["name"]})

            if 'hash_mode' in pap and 'lacp' not in pap:
                if not pap_db.lacp and pap['hash_mode'].lower() != 'l2':
                    raise ext_pap.InvalidLacpValue(hash_mode=pap["hash_mode"])
                pap_db.update({"hash_mode": pap["hash_mode"]})

            if 'lacp' in pap and 'hash_mode' not in pap:
                lacp_flag = self._lacp_flag(pap)
                if not lacp_flag and pap_db.hash_mode != "L2":
                    raise ext_pap.InvalidLacpValue(hash_mode=pap_db.hash_mode)
                pap_db.update({"lacp": self._lacp_flag(pap)})

            if 'lacp' in pap and 'active_standby' not in pap:
                lacp_flag = self._lacp_flag(pap)
                if lacp_flag and pap_db.active_standby:
                    raise ext_pap.InvalidAsValueLacp
                pap_db.update({"lacp": lacp_flag})

            if 'lacp' in pap and 'active_standby' in pap:
                lacp_flag = self._lacp_flag(pap)
                as_flag = self._as_flag(pap)
                if lacp_flag and as_flag:
                    raise ext_pap.InvalidAsValueLacp
                pap_db.update({"lacp": lacp_flag})

            if 'hash_mode' in pap and 'lacp' in pap:
                if not pap['lacp'] and pap['hash_mode'].lower() != 'l2':
                    raise ext_pap.InvalidLacpValue(hash_mode=pap["hash_mode"])
                pap_db.update({"hash_mode": pap["hash_mode"]})
                pap_db.update({"lacp": self._lacp_flag(pap)})

            if 'active_standby' in pap and "lacp" not in pap:
                as_flag = self._as_flag(pap)
                if pap_db.lacp and as_flag:
                    raise ext_pap.InvalidLacpValueAs
                pap_db.update({"active_standby": as_flag})

            if 'active_standby' in pap and "lacp" in pap:
                as_flag = self._as_flag(pap)
                lacp_flag = self._lacp_flag(pap)
                if lacp_flag and as_flag:
                    raise ext_pap.InvalidLacpValueAs
                pap_db.update({"active_standby": as_flag})

            if 'add_interfaces' in pap:
                self._check_ifc(context, pap["add_interfaces"])
                for interface in pap["add_interfaces"]:
                    interface_db = Interface(
                                   hostname=interface["hostname"],
                                   interface=interface["interface"],
                                   pap=pap_db)
                    context.session.add(interface_db)
            if 'remove_interfaces' in pap:
                for interface in pap["remove_interfaces"]:
                    query = self._model_query(context, Interface)
                    query.filter_by(hostname=interface["hostname"],
                          interface=interface["interface"]).delete()
            query = self._model_query(context, Interface)
            pap_db.interfaces = query.filter_by(pap_id=pap_id).all()
        return self._make_pap_dict(pap_db)

    def get_physical_attachment_point(self, context, pap_id, fields=None):
        """
        Gets an existing physical attachment point
        Args:
             pap_id = uuid of the physical attachment point being requested
        """

        try:
            query = self._model_query(context, PhysicalAttachmentPoint)
            pap_db = query.filter_by(id=pap_id).one()
        except exc.NoResultFound:
            raise ext_pap.NoPhysicalAttachmentPointFound(id=pap_id)
        return self._make_pap_dict(pap_db, fields)

    def get_physical_attachment_points(self, context, filters=None,
             fields=None, sorts=None, limit=None, marker=None,
             page_reverse=None):
        """
        Gets the list of all the existing physical attachment points
        """
        return self._get_collection(context, PhysicalAttachmentPoint,
                                    self._make_pap_dict, filters=filters,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker, fields=fields,
                                    page_reverse=page_reverse)

    def delete_physical_attachment_point(self, context, pap_id):
        """
        Deletes an existing physical attachment point
        Args:
             pap_id = uuid of the physical attachment point being deleted
        """

        try:
            query = context.session.query(PhysicalAttachmentPoint)
            pap_db = query.filter_by(id=pap_id).first()
        except exc.NoResultFound:
            raise ext_pap.NoPhysicalAttachmentPointFound(id=pap_id)
        with context.session.begin(subtransactions=True):
            context.session.delete(pap_db)

    def _make_pap_dict(self, pap, fields=None):
        interfaces = []
        for interface in pap.interfaces:
            interfaces.append({"hostname": interface.hostname,
                               "interface": interface.interface})

        pap_dict = {"id": pap.id,
                    "name": pap.name,
                    "interfaces": interfaces,
                    "hash_mode": pap.hash_mode,
                    "lacp": pap.lacp,
                    "active_standby": pap.active_standby,
                    "tenant_id": pap.tenant_id,
                    "transit_domain_id": pap.transit_domain_id,
                    "implicit": pap.implicit}
        return self._fields(pap_dict, fields)

    def _lacp_flag(self, pap):
        if str(pap.get("lacp")).lower() == 'false':
            return False
        elif str(pap.get("lacp")).lower() == 'true':
            return True
        else:
            raise Exception("Please pass True/true or False/false for LACP")

    def _as_flag(self, pap):
        if str(pap.get("active_standby")).lower() == 'false':
            return False
        elif str(pap.get("active_standby")).lower() == 'true':
            return True
        else:
            raise Exception("Please pass True/true or False/false"
                            "for active_standby")

    def _check_transit_domain_limit(self, context, td_id):
        try:
            query = self._model_query(context, PhysicalAttachmentPoint)
            pap_db = query.filter_by(transit_domain_id=td_id).one()
            if pap_db:
                raise ext_pap.TransitDomainLimit
        except exc.NoResultFound:
            return False

    def _check_ifc(self, context, interfaces):
        for interface in interfaces:
            query = self._model_query(context, Interface)
            host = interface["hostname"]
            ifc_name = interface["interface"]
            try:
                ifc = query.filter_by(hostname=host, interface=ifc_name).one()
                if ifc:
                    raise ext_pap.InterfaceInUse(ifc=host + "+" + ifc_name,
                              id=ifc.pap_id)
            except exc.NoResultFound:
                pass



# Copyright 2015 PLUMgrid, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from networking_plumgrid._i18n import _
from neutron.api.v2 import base
from neutron.common import exceptions

from webob import exc as web_exc


class L2GatewayInUse(exceptions.InUse):
    message = _("L2 Gateway '%(gateway_id)s' still has active mappings "
                "with one or more neutron networks.")


class L2GatewayNotFound(exceptions.NotFound):
    message = _("L2 Gateway %(gateway_id)s could not be found")


class L2GatewayDeviceInUse(exceptions.InUse):
    message = _("L2 Gateway Device '%(device_id)s' is still used by "
                "one or more network gateways.")


class L2GatewayDeviceNotFound(exceptions.NotFound):
    message = _("L2 Gateway Device %(device_id)s could not be found.")


class L2GatewayDeviceNameNotFound(exceptions.NotFound):
    message = _("L2 Gateway Device %(device_name)s could not be found.")


class L2GatewayPortInUse(exceptions.InUse):
    message = _("Port '%(port_id)s' is owned by '%(device_owner)s' and "
                "therefore cannot be deleted directly via the port API.")


class L2GatewayConnectionExists(exceptions.InUse):
    message = _("The specified mapping '%(mapping)s' exists on "
                "network gateway '%(gateway_id)s'.")


class L2MultipleGatewayConnections(exceptions.NeutronException):
    message = _("Multiple network connections found on '%(gateway_id)s' "
                "with provided criteria.")


class L2GatewayInterfaceNotFound(exceptions.NeutronException):
    message = _("L2 Gateway interface '%(interface_id)s' not found.")


class L2GatewayPhysicalPortFaultStatus(exceptions.NeutronException):
    message = _("Physical Port '%(int_name)s' in Physical switch "
                "'%(device_name)s' port_fault_status is '%(fault_status)s'")


class L2GatewayPhysicalSwitchFaultStatus(exceptions.NeutronException):
    message = _("Physical Switch '%(device_name)s' switch_fault_status is "
                "'%(fault_status)s'")


class L2GatewayConnectionNotFound(exceptions.NotFound):
    message = _("The connection %(id)s not found on the l2 gateway")


class L2gatewaySegmentationIDNotFound(exceptions.NotFound):
    message = _("The segmentation id is not found on gateway "
                "'%(gateway_id)s'")


class MultipleSegmentsFound(exceptions.NeutronException):
    message = _("Multiple segments found for the network  '%(network_id)s'")


class L2GatewayDeviceRequired(exceptions.NeutronException):
    message = _("L2 Gateway Device is required")


class L2GatewayInterfaceRequired(exceptions.NeutronException):
    message = _("L2 Gateway Interface required")


class L2GatewaySegmentationIDExists(exceptions.InUse):
    message = _("The segmentation id is already specified in "
                "gateway interface '%(interface_id)s'.")


class L2GatewaySegmentationRequired(exceptions.NeutronException):
    message = _("L2 gateway segmentation id is required and must be "
                "consistent for all the interfaces")


class L2GatewayDuplicateSegmentationID(exceptions.Conflict):
    message = _("%(message)s")


class OVSDBError(exceptions.NeutronException):
    message = _("%(message)s")

base.FAULT_MAP.update({L2GatewayInUse: web_exc.HTTPConflict,
                       L2GatewayPortInUse: web_exc.HTTPConflict,
                       L2GatewayConnectionExists: web_exc.HTTPConflict,
                       L2GatewayConnectionNotFound: web_exc.HTTPNotFound,
                       MultipleSegmentsFound: web_exc.HTTPConflict,
                       L2GatewaySegmentationRequired: web_exc.HTTPConflict,
                       L2MultipleGatewayConnections: web_exc.HTTPConflict,
                       L2GatewayDuplicateSegmentationID: web_exc.HTTPConflict,
                       OVSDBError: web_exc.HTTPConflict})



# Copyright 2016 PLUMgrid, Inc. All Rights Reserved.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import mock
from networking_plumgrid.neutron.tests.unit.cli import test_cli20
from networking_plumgrid.neutronclient.policy import (
    policy_service as ps)
from neutronclient import shell
import sys


class CLITestV20ExtensionPolicyServiceJSON(test_cli20.CLITestV20Base):
    def setUp(self):
        # need to mock before super because extensions loaded on instantiation
        self._mock_extension_loading()
        super(CLITestV20ExtensionPolicyServiceJSON,
              self).setUp(plurals={'tags': 'tag'})

    def _mock_extension_loading(self):
        ext_pkg = 'neutronclient.common.extension'
        contrib = mock.patch(ext_pkg + '._discover_via_entry_points').start()
        contrib.return_value = [("policy_service", ps)]
        return contrib

    def test_ext_cmd_loaded(self):
        """Tests policy service commands loaded."""
        shell.NeutronShell('2.0')
        ext_cmd = {'policy-service-list':
                   ps.PolicyServiceList,
                   'policy-service-create':
                   ps.PolicyServiceCreate,
                   'policy-service-update':
                   ps.PolicyServiceUpdate,
                   'policy-service-delete':
                   ps.PolicyServiceDelete,
                   'policy-service-show':
                   ps.PolicyServiceShow}
        self.assertDictContainsSubset(ext_cmd, shell.COMMANDS['2.0'])

    def test_create_policy_service_name(self):
        # Create policy service: ps1.
        resource = 'policy_service'
        cmd = ps.PolicyServiceCreate(test_cli20.MyApp(
                                      sys.stdout), None)
        name = 'ps1'
        args = [name]
        position_names = ['name']
        position_values = [name]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_policy_service_desc(self):
        # Create policy service: ps1.
        resource = 'policy_service'
        cmd = ps.PolicyServiceCreate(test_cli20.MyApp(
                                      sys.stdout), None)
        name = 'ps1'
        desc = 'policy service'
        args = [name, '--description', desc]
        position_names = ['name', 'description']
        position_values = [name, desc]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_policy_service_inport(self):
        # Create policy service: ps1.
        resource = 'policy_service'
        cmd = ps.PolicyServiceCreate(test_cli20.MyApp(
                                      sys.stdout), None)
        name = 'ps1'
        inport = [{'id': 'portid'}]
        args = [name, '--ingress-port', 'id=portid']
        position_names = ['name', 'ingress_ports']
        position_values = [name, inport]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_policy_service_eport(self):
        # Create policy service: ps1.
        resource = 'policy_service'
        cmd = ps.PolicyServiceCreate(test_cli20.MyApp(
                                      sys.stdout), None)
        name = 'ps1'
        eport = [{'id': 'portid'}]
        args = [name, '--egress-port', 'id=portid']
        position_names = ['name', 'egress_ports']
        position_values = [name, eport]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_policy_service_bidirect_port(self):
        # Create policy service: ps1.
        resource = 'policy_service'
        cmd = ps.PolicyServiceCreate(test_cli20.MyApp(
                                      sys.stdout), None)
        name = 'ps1'
        bidirect_port = [{'id': 'portid'}]
        args = [name, '--bidirectional-port', 'id=portid']
        position_names = ['name', 'bidirectional_ports']
        position_values = [name, bidirect_port]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_policy_service_inport_name(self):
        # Create policy service: ps1.
        resource = 'policy_service'
        cmd = ps.PolicyServiceCreate(test_cli20.MyApp(
                                      sys.stdout), None)
        name = 'ps1'
        inport = [{'id': 'portid'}]
        args = [name, '--ingress-port', 'name=portid']
        position_names = ['name', 'ingress_ports']
        position_values = [name, inport]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_policy_service_eport_name(self):
        # Create policy service: ps1.
        resource = 'policy_service'
        cmd = ps.PolicyServiceCreate(test_cli20.MyApp(
                                      sys.stdout), None)
        name = 'ps1'
        eport = [{'id': 'portid'}]
        args = [name, '--egress-port', 'name=portid']
        position_names = ['name', 'egress_ports']
        position_values = [name, eport]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_policy_service_bidirect_port_name(self):
        # Create policy service: ps1.
        resource = 'policy_service'
        cmd = ps.PolicyServiceCreate(test_cli20.MyApp(
                                      sys.stdout), None)
        name = 'ps1'
        bidirect_port = [{'id': 'portid'}]
        args = [name, '--bidirectional-port', 'name=portid']
        position_names = ['name', 'bidirectional_ports']
        position_values = [name, bidirect_port]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_policy_service_inport_eport(self):
        # Create policy service: ps1.
        resource = 'policy_service'
        cmd = ps.PolicyServiceCreate(test_cli20.MyApp(
                                      sys.stdout), None)
        name = 'ps1'
        inport = [{'id': 'portid'}]
        eport = [{'id': 'portid'}]
        args = [name, '--ingress-port', 'id=portid',
                '--egress-port', 'id=portid']
        position_names = ['name', 'ingress_ports', 'egress_ports']
        position_values = [name, inport, eport]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_update_policy_service_name(self):
        # Update policy service: myid --name
        resource = 'policy_service'
        cmd = ps.PolicyServiceUpdate(test_cli20.MyApp(
                                     sys.stdout), None)
        args = ['myid', '--name', 'myps']
        values = {'name': 'myps'}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_update_policy_service_desc(self):
        # Update policy service: myid --description
        resource = 'policy_service'
        cmd = ps.PolicyServiceUpdate(test_cli20.MyApp(
                                     sys.stdout), None)
        args = ['myid', '--description', 'myps desc']
        values = {'description': 'myps desc'}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_update_policy_service_add_inport(self):
        # Update policy service: myid --add-ingress-port
        resource = 'policy_service'
        cmd = ps.PolicyServiceUpdate(test_cli20.MyApp(
                                     sys.stdout), None)
        args = ['myid', '--add-ingress-port', 'id=portid']
        values = {'add_ingress_ports': [{'id': 'portid'}]}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_update_policy_service_add_eport(self):
        # Update policy service: myid --add-egress-port
        resource = 'policy_service'
        cmd = ps.PolicyServiceUpdate(test_cli20.MyApp(
                                     sys.stdout), None)
        args = ['myid', '--add-egress-port', 'id=portid']
        values = {'add_egress_ports': [{'id': 'portid'}]}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_update_policy_service_remove_inport(self):
        # Update policy service: myid --remove-ingress-port
        resource = 'policy_service'
        cmd = ps.PolicyServiceUpdate(test_cli20.MyApp(
                                     sys.stdout), None)
        args = ['myid', '--remove-ingress-port', 'id=portid']
        values = {'remove_ingress_ports': [{'id': 'portid'}]}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_update_policy_service_remove_eport(self):
        # Update policy service: myid --remove-egress-port
        resource = 'policy_service'
        cmd = ps.PolicyServiceUpdate(test_cli20.MyApp(
                                     sys.stdout), None)
        args = ['myid', '--remove-egress-port', 'id=portid']
        values = {'remove_egress_ports': [{'id': 'portid'}]}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_update_policy_service_add_bidirect_port(self):
        # Update policy service: myid --add-bidirectional-port
        resource = 'policy_service'
        cmd = ps.PolicyServiceUpdate(test_cli20.MyApp(
                                     sys.stdout), None)
        args = ['myid', '--add-bidirectional-port', 'id=portid']
        values = {'add_bidirectional_ports': [{'id': 'portid'}]}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_update_policy_service_remove_bidirect_port(self):
        # Update policy service: myid --remove-bidirectional-port
        resource = 'policy_service'
        cmd = ps.PolicyServiceUpdate(test_cli20.MyApp(
                                     sys.stdout), None)
        args = ['myid', '--remove-bidirectional-port', 'id=portid']
        values = {'remove_bidirectional_ports': [{'id': 'portid'}]}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_list_policy_services(self):
        # list policy services: -D.
        resources = "policy_services"
        cmd = ps.PolicyServiceList(test_cli20.MyApp(
                                   sys.stdout), None)
        self._test_list_resources(resources, cmd, True)

    def test_list_policy_services_sort(self):
        # list policy services:
        # --sort-key name --sort-key id --sort-key asc --sort-key desc
        resources = "policy_services"
        cmd = ps.PolicyServiceList(test_cli20.MyApp(
                                   sys.stdout), None)
        self._test_list_resources(resources, cmd,
                                  sort_key=["name", "id"],
                                  sort_dir=["asc", "desc"])

    def test_list_policy_services_limit(self):
        # list policy services: -P.
        resources = "policy_services"
        cmd = ps.PolicyServiceList(test_cli20.MyApp(
                                   sys.stdout), None)
        self._test_list_resources(resources, cmd, page_size=1000)

    def test_delete_policy_service(self):
        # Delete policy service: myid.
        resource = "policy_service"
        cmd = ps.PolicyServiceDelete(test_cli20.MyApp(
                                sys.stdout), None)
        my_id = 'myid'
        args = [my_id]
        self._test_delete_resource(resource, cmd, my_id, args)

    def test_show_policy_service(self):
        # Show policy service: --fields id.
        resource = 'policy_service'
        cmd = ps.PolicyServiceShow(test_cli20.MyApp(
                              sys.stdout), None)
        args = ['--fields', 'id', '--fields', 'name', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id, args,
                                 ['id', 'name'])



# Copyright 2016 PLUMgrid, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import networking_plumgrid
from networking_plumgrid._i18n import _
from neutron.api import extensions
from neutron.api.v2 import attributes as attr
from neutron.api.v2 import base
from neutron.common import exceptions as nexceptions
from neutron import manager

PG_POLICY_TAG = 'policy_tag'
PG_POLICY_TAGS = 'policy-tags'

extensions.append_api_extensions_path(
    networking_plumgrid.neutron.plugins.extensions.__path__)


class NoPolicyTagFound(nexceptions.NotFound):
    message = _("Policy Tag with id %(id)s does not exist")


class UpdateParametersRequired(nexceptions.InvalidInput):
    message = _("No update parameter specified atleast one needed")


class InvalidPolicyTagId(nexceptions.InvalidInput):
    message = _("Invalid input for Policy Tag ID. Please specify Tag ID "
                "in range 257-2047")


def _validate_tag_id(data, valid_values=None):
    if not data:
        return
    try:
        tag = int(data)
    except (ValueError, TypeError):
        raise InvalidPolicyTagId()
    if not (tag >= 257 and tag <= 2047):
        raise InvalidPolicyTagId()

attr.validators['type:validate_tag_id'] = _validate_tag_id
pt_supported_types = ['fip', 'dot1q', 'nsh']

RESOURCE_ATTRIBUTE_MAP = {
    'policy_tag': {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {
                'type:uuid': None
            },
            'is_visible': True,
            'primary_key': True
        },
        'name': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'default': ''
        },
        'tag_type': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'default': '',
            'validate': {'type:values': pt_supported_types}
        },
        'tag_id': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'default': None,
            'validate': {
                'type:validate_tag_id': None
            }
        },
        'floatingip_id': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'default': None,
            'validate': {
                'type:uuid_or_none': None
            },
        },
        'floating_ip_address': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'default': ''
        },
        'router_id': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'default': None,
        },
        'tenant_id': {'allow_post': True,
                      'allow_put': False,
                      'required_by_policy': True,
                      'validate': {'type:string': attr.TENANT_ID_MAX_LEN},
                      'is_visible': True},
    }
}


class Policytag(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Policy Tag"

    @classmethod
    def get_alias(cls):
        return "policy-tag"

    @classmethod
    def get_description(cls):
        return "This API will be used to create policy tags" \
               " from Neutron and map it in PLUMgrid."

    @classmethod
    def get_namespace(cls):
        return "http://docs.openstack.org/ext/policy_tag" \
               "/api/v2.0"

    @classmethod
    def get_updated(cls):
        return "2016-01-01T00:00:00-00:00"

    @classmethod
    def get_resources(cls):
        exts = []
        plugin = manager.NeutronManager.get_plugin()

        resource_name = PG_POLICY_TAG
        collection_name = PG_POLICY_TAGS
        params = RESOURCE_ATTRIBUTE_MAP.get(resource_name, dict())
        controller = base.create_resource(
            collection_name, resource_name, plugin, params)
        ex = extensions.ResourceExtension(collection_name,
                                          controller)
        exts.append(ex)
        return exts



# Copyright 2016 PLUMgrid, Inc. All Rights Reserved.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import mock
from networking_plumgrid.neutron.tests.unit.cli import test_cli20
from networking_plumgrid.neutronclient.policy import (
    policy_tag as ptag)
from neutronclient import shell
import sys


class CLITestV20ExtensionPolicyTagJSON(test_cli20.CLITestV20Base):
    def setUp(self):
        # need to mock before super because extensions loaded on instantiation
        self._mock_extension_loading()
        super(CLITestV20ExtensionPolicyTagJSON,
              self).setUp(plurals={'tags': 'tag'})

    def _mock_extension_loading(self):
        ext_pkg = 'neutronclient.common.extension'
        contrib = mock.patch(ext_pkg + '._discover_via_entry_points').start()
        contrib.return_value = [("policy_tag", ptag)]
        return contrib

    def test_ext_cmd_loaded(self):
        """Tests policy tag commands loaded."""
        shell.NeutronShell('2.0')
        ext_cmd = {'policy-tag-list':
                   ptag.PolicyTagList,
                   'policy-tag-create':
                   ptag.PolicyTagCreate,
                   'policy-tag-update':
                   ptag.PolicyTagUpdate,
                   'policy-tag-delete':
                   ptag.PolicyTagDelete,
                   'policy-tag-show':
                   ptag.PolicyTagShow}
        self.assertDictContainsSubset(ext_cmd, shell.COMMANDS['2.0'])

    def test_create_policy_tag_name(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        args = [name]
        position_names = ['name']
        position_values = [name]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_policy_tag_desc(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        description = 'ptag1:desc'
        args = [name, '--description', description]
        position_names = ['name', 'description']
        position_values = [name, description]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_policy_tag_type_nsh(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        ptag_type = 'nsh'
        args = [name, '--type', ptag_type]
        position_names = ['name', 'tag_type']
        position_values = [name, ptag_type]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_policy_tag_type_dot1q(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        ptag_type = 'dot1q'
        args = [name, '--type', ptag_type]
        position_names = ['name', 'tag_type']
        position_values = [name, ptag_type]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_policy_tag_type_floatingip(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        ptag_type = 'fip'
        args = [name, '--type', ptag_type]
        position_names = ['name', 'tag_type']
        position_values = [name, ptag_type]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_policy_tag(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        ptag_type = 'nsh'
        tag_id = '20'
        args = [name, '--type', ptag_type, '--tag-id', tag_id]
        position_names = ['name', 'tag_type', 'tag_id']
        position_values = [name, ptag_type, tag_id]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_policy_tag_floatingip(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        ptag_type = 'fip'
        floatingip = 'fipid'
        args = [name, '--type', ptag_type, '--floating-ip', floatingip]
        position_names = ['name', 'tag_type', 'floatingip_id']
        position_values = [name, ptag_type, floatingip]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_policy_tag_type_fip_tag_id(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        ptag_type = 'fip'
        floatingip = 'fipid'
        tag_id = '20'
        args = [name, '--type', ptag_type, '--floating-ip', floatingip,
                '--tag-id', tag_id]
        position_names = ['name', 'tag_type', 'floatingip_id', 'tag_id']
        position_values = [name, ptag_type, floatingip, tag_id]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_policy_tag_type_nsh_fip(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        ptag_type = 'nsh'
        floatingip = 'fipid'
        tag_id = '20'
        args = [name, '--type', ptag_type, '--floating-ip', floatingip,
                '--tag-id', tag_id]
        position_names = ['name', 'tag_type', 'floatingip_id', 'tag_id']
        position_values = [name, ptag_type, floatingip, tag_id]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_policy_tag_type_dot1q_fip(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        ptag_type = 'dot1q'
        floatingip = 'fipid'
        tag_id = '20'
        args = [name, '--type', ptag_type, '--floating-ip', floatingip,
                '--tag-id', tag_id]
        position_names = ['name', 'tag_type', 'floatingip_id', 'tag_id']
        position_values = [name, ptag_type, floatingip, tag_id]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_policy_tag_router_id(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        ptag_type = 'fip'
        floatingip = 'fipid'
        routerid = 'routerid'
        args = [name, '--type', ptag_type, '--floating-ip', floatingip,
                '--router-id', routerid]
        position_names = ['name', 'tag_type', 'floatingip_id', 'router_id']
        position_values = [name, ptag_type, floatingip, routerid]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_policy_tag_invalid_tag_type(self):
        # Create policy tag: ptag1.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagCreate(test_cli20.MyApp(
                                   sys.stdout), None)
        name = 'ptag1'
        ptag_type = 'invalid'
        args = [name, '--type', ptag_type]
        position_names = ['name', 'tag_type']
        position_values = [name, ptag_type]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_list_policy_tags(self):
        # list policytags: -D.
        resources = "policy_tags"
        cmd = ptag.PolicyTagList(test_cli20.MyApp(
                                 sys.stdout), None)
        self._test_list_resources(resources, cmd, True)

    def test_list_policy_tags_sort(self):
        # list policytags:
        # --sort-key name --sort-key id --sort-key asc --sort-key desc
        resources = "policy_tags"
        cmd = ptag.PolicyTagList(test_cli20.MyApp(
                                 sys.stdout), None)
        self._test_list_resources(resources, cmd,
                                  sort_key=["name", "id"],
                                  sort_dir=["asc", "desc"])

    def test_list_policy_tags_limit(self):
        # list policytags: -P.
        resources = "policy_tags"
        cmd = ptag.PolicyTagList(test_cli20.MyApp(
                                 sys.stdout), None)
        self._test_list_resources(resources, cmd, page_size=1000)

    def test_delete_policy_tag(self):
        # Delete policytag: myid.
        resource = "policy_tag"
        cmd = ptag.PolicyTagDelete(test_cli20.MyApp(
                                   sys.stdout), None)
        my_id = 'myid'
        args = [my_id]
        self._test_delete_resource(resource, cmd, my_id, args)

    def test_show_policy_tag(self):
        # Show policytag: --fields id.
        resource = 'policy_tag'
        cmd = ptag.PolicyTagShow(test_cli20.MyApp(
                                 sys.stdout), None)
        args = ['--fields', 'id', '--fields', 'name', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id, args,
                                 ['id', 'name'])

    def test_update_policy_tag_name(self):
        # Update policytag: myid --name
        resource = 'policy_tag'
        cmd = ptag.PolicyTagUpdate(test_cli20.MyApp(
                                   sys.stdout), None)
        args = ['myid', '--name', 'myptag']
        values = {'name': 'myptag'}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)



# Copyright 2015 PLUMgrid, Inc. All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from mox3 import mox
from neutronclient.neutron import v2_0
from neutronclient import shell as neutronshell
from neutronclient.tests.unit import test_cli20 as neutron_test_cli20
from neutronclient.v2_0 import client

TOKEN = neutron_test_cli20.TOKEN
end_url = neutron_test_cli20.end_url


class MyResp(neutron_test_cli20.MyResp):
    """
    Class to store response information
    """
    pass


class MyApp(neutron_test_cli20.MyApp):
    """
    Application Class
    """
    pass


class MyComparator(neutron_test_cli20.MyComparator):
    """
    Class for performing comparison operations
    """
    pass


class CLITestV20Base(neutron_test_cli20.CLITestV20Base):
    """
    Base Class for CLI Tests
    """

    def setUp(self, plurals=None):
        super(CLITestV20Base, self).setUp()
        self.client = client.Client(token=TOKEN,
                                    endpoint_url=self.endurl)

    def _test_create_resource(self, resource, cmd, name, myid, args,
                              position_names, position_values,
                              tenant_id=None, tags=None, admin_state_up=True,
                              extra_body=None, cmd_resource=None,
                              parent_id=None, **kwargs):
        self.mox.StubOutWithMock(cmd, "get_client")
        self.mox.StubOutWithMock(self.client.httpclient, "request")
        cmd.get_client().MultipleTimes().AndReturn(self.client)
        if not cmd_resource:
            cmd_resource = resource
        body = {resource: {}, }
        body[resource].update(kwargs)

        for i in range(len(position_names)):
            body[resource].update({position_names[i]: position_values[i]})
        ress = {resource:
                {self.id_field: myid}, }
        if name:
            ress[resource].update({'name': name})
        self.client.format = self.format
        resstr = self.client.serialize(ress)
        # url method body to retrieve resource plural from v2_0
        resource_plural = v2_0._get_resource_plural(cmd_resource, self.client)
        path = getattr(self.client, resource_plural + "_path")
        if self.format == 'json':
            mox_body = MyComparator(body, self.client)
        else:
            mox_body = self.client.serialize(body)
        self.client.httpclient.request(
            end_url(path, format=self.format), 'POST',
            body=mox_body,
            headers=mox.ContainsKeyValue(
                'X-Auth-Token', TOKEN)).AndReturn((MyResp(200), resstr))
        args.extend(['--request-format', self.format])
        self.mox.ReplayAll()
        cmd_parser = cmd.get_parser('create_' + resource)
        neutronshell.run_command(cmd, cmd_parser, args)
        self.mox.VerifyAll()
        self.mox.UnsetStubs()
        _str = self.fake_stdout.make_string()
        self.assertIn(myid, _str)
        if name:
            self.assertIn(name, _str)



# Copyright 2015 PLUMgrid, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""policy

Revision ID: 1ae50f04cdd5
Revises: 351c4f5710e7
Create Date: 2016-07-29 15:35:12.194948

"""

# revision identifiers, used by Alembic.
revision = '1ae50f04cdd5'
down_revision = '351c4f5710e7'

from alembic import op
import sqlalchemy as sa

tag_enum = sa.Enum('dot1q', 'fip', 'nsh',
                   name='pg_policy_tags_type')
direction_enum = sa.Enum('ingress', 'egress', 'bidirectional',
                         name='pg_port_group_direction')
policy_rule_protocol_enum = sa.Enum('any', 'tcp', 'udp', 'icmp',
                                    name='pg_policy_rule_protocol')
policy_action_enum = sa.Enum('allow', 'copy',
                             name='pg_policy_rule_action')


def upgrade():
    op.create_table(
        'pg_policy_tags',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('tag_id', sa.String(length=1024), nullable=True),
        sa.Column('tag_type', tag_enum, nullable=True),
        sa.Column('floatingip_address', sa.String(length=255), nullable=True),
        sa.Column('floatingip_id', sa.String(length=36), nullable=True),
        sa.Column('router_id', sa.String(length=36), nullable=True),
        sa.Column('tenant_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['floatingip_id'], ['floatingips.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'pg_endpoint_groups',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('description', sa.String(length=1024), nullable=True),
        sa.Column('policy_tag_id', sa.String(length=36), nullable=True),
        sa.Column('tenant_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['policy_tag_id'], ['pg_policy_tags.id'],
                                ondelete='RESTRICT'),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'pg_security_policy_tag_binding',
        sa.Column('security_group_id', sa.String(length=36), nullable=False),
        sa.Column('policy_tag_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['security_group_id'], ['securitygroups.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['policy_tag_id'],
                                ['pg_policy_tags.id'])
    )

    op.create_table(
        'pg_endpoints',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('ip_mask', sa.String(length=36), nullable=True),
        sa.Column('port_id', sa.String(length=36), nullable=True),
        sa.Column('ip_port', sa.String(length=255), nullable=True),
        sa.Column('tenant_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['port_id'], ['ports.id'], ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'pg_endpoint_group_member_binding',
        sa.Column('endpoint_id', sa.String(length=36), nullable=False),
        sa.Column('endpoint_group_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['endpoint_id'], ['pg_endpoints.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['endpoint_group_id'],
                                ['pg_endpoint_groups.id'], ondelete='CASCADE')
    )

    op.create_table(
        'pg_policy_services',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('description', sa.String(length=1024), nullable=True),
        sa.Column('tenant_id', sa.String(length=36), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'pg_policy_service_port_binding',
        sa.Column('policy_service_id', sa.String(length=36), nullable=False),
        sa.Column('port_id', sa.String(length=36), nullable=False),
        sa.Column('direction', direction_enum, nullable=False),
        sa.ForeignKeyConstraint(['policy_service_id'],
                                ['pg_policy_services.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['port_id'],
                                ['ports.id'], ondelete='CASCADE')
    )

    op.create_table(
        'pg_policy_rules',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('src_grp_epg', sa.String(length=36), nullable=True),
        sa.Column('src_grp_sg', sa.String(length=36), nullable=True),
        sa.Column('dst_grp_epg', sa.String(length=36), nullable=True),
        sa.Column('dst_grp_sg', sa.String(length=36), nullable=True),
        sa.Column('protocol', policy_rule_protocol_enum, nullable=False),
        sa.Column('src_port_range', sa.String(length=255), nullable=True),
        sa.Column('dst_port_range', sa.String(length=255), nullable=True),
        sa.Column('action', policy_action_enum, nullable=False),
        sa.Column('action_target_service', sa.String(length=36),
                  nullable=True),
        sa.Column('action_target_tenant_id', sa.String(length=36),
                  nullable=True),
        sa.Column('tag', sa.String(length=36), nullable=True),
        sa.Column('tenant_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['src_grp_epg'], ['pg_endpoint_groups.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['src_grp_sg'], ['securitygroups.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['dst_grp_epg'], ['pg_endpoint_groups.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['dst_grp_sg'], ['securitygroups.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['action_target_service'],
                                ['pg_policy_services.id'],
                                ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['tag'], ['pg_policy_tags.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id')
    )



# Copyright 2015 PLUMgrid, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_plumgrid.neutron.plugins.db import pg_models


def add_network_binding(session, network_id, network_type, physical_network,
                        vlan_id):
    binding = pg_models.ProviderNetBinding(network_id=network_id,
                                           network_type=network_type,
                                           physical_network=physical_network,
                                           vlan_id=vlan_id)
    session.add(binding)
    return binding


def get_network_binding(session, network_id):
    qry = session.query(pg_models.ProviderNetBinding)
    nets = qry.filter_by(network_id=network_id).first()
    return nets



# Copyright 2016 PLUMgrid, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Endpoint Group extension unit tests
"""
from networking_plumgrid.neutron.plugins.common import \
    exceptions as p_excep
from networking_plumgrid.neutron.plugins.extensions import \
    endpointgroup
from networking_plumgrid.neutron.tests.unit import \
    test_networking_plumgrid as test_pg

from neutron.api import extensions
from neutron import context
from neutron import manager
from neutron.tests.unit.api import test_extensions as test_ext
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class EndpointGroupExtensionManager(object):

    def get_resources(self):
        return endpointgroup.Endpointgroup.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class EndpointGroupTestCase(test_pg.PLUMgridPluginV2TestCase):
    def setUp(self, plugin=None, ext_mgr=None):
        super(EndpointGroupTestCase, self).setUp()
        ext_mgr = EndpointGroupExtensionManager()
        extensions.PluginAwareExtensionManager._instance = None
        self.ext_api = test_ext.setup_extensions_middleware(ext_mgr)


class TestEndpointGroup(EndpointGroupTestCase):
    def test_create_endpoint_group(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        epg = self._make_epg_dict()

        epg_ret = plugin.create_endpoint_group(
                      admin_context, epg)
        epg["endpoint_group"]["id"] = epg_ret["id"]
        self.assertEqual(epg_ret, epg["endpoint_group"])

    def test_create_endpoint_group_with_policy_tag(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ptag = self._fake_policy_tag_dict()
        ptag_ret = plugin.create_policy_tag(admin_context, ptag)
        ptag["policy_tag"]["id"] = ptag_ret["id"]

        epg = self._make_epg_dict(ptag_id=ptag_ret["id"])

        epg_ret = plugin.create_endpoint_group(
                      admin_context, epg)
        epg["endpoint_group"]["id"] = epg_ret["id"]
        self.assertEqual(epg_ret, epg["endpoint_group"])

    def test_create_endpoint_group_with_invalid_policy_tag(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ptag = self._fake_policy_tag_dict()
        ptag_ret = plugin.create_policy_tag(admin_context, ptag)
        ptag["policy_tag"]["id"] = ptag_ret["id"]

        epg = self._make_epg_dict(ptag_id="abcdef123")

        self.assertRaises(p_excep.PLUMgridException,
                          plugin.create_endpoint_group, admin_context, epg)

    def test_create_endpoint_group_with_in_use_policy_tag(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ptag = self._fake_policy_tag_dict()
        ptag_ret = plugin.create_policy_tag(admin_context, ptag)
        ptag["policy_tag"]["id"] = ptag_ret["id"]

        # Create first endpoint group with policy tag
        epg1 = self._make_epg_dict(ptag_id=ptag_ret["id"])
        epg_ret1 = plugin.create_endpoint_group(
                      admin_context, epg1)
        epg1["endpoint_group"]["id"] = epg_ret1["id"]

        # Create a second endpoint group with same policy tag
        epg2 = self._make_epg_dict(ptag_id=ptag_ret["id"])
        epg_ret2 = plugin.create_endpoint_group(
                      admin_context, epg2)
        epg2["endpoint_group"]["id"] = epg_ret2["id"]
        self.assertEqual(epg_ret1, epg1["endpoint_group"])
        self.assertEqual(epg_ret2, epg2["endpoint_group"])

    def test_create_delete_endpoint_group(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ptag = self._fake_policy_tag_dict()
        ptag_ret = plugin.create_policy_tag(admin_context, ptag)
        ptag["policy_tag"]["id"] = ptag_ret["id"]

        epg = self._make_epg_dict(ptag_id=ptag_ret["id"])

        epg_ret = plugin.create_endpoint_group(
                      admin_context, epg)
        epg["endpoint_group"]["id"] = epg_ret["id"]

        resp = plugin.delete_endpoint_group(admin_context, epg_ret["id"])
        self.assertEqual(None, resp)

    def test_create_update_endpoint_group(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ptag = self._fake_policy_tag_dict()
        ptag_ret = plugin.create_policy_tag(admin_context, ptag)
        ptag["policy_tag"]["id"] = ptag_ret["id"]

        epg = self._make_epg_dict()

        epg_ret = plugin.create_endpoint_group(
                      admin_context, epg)
        epg["endpoint_group"]["id"] = epg_ret["id"]

        epg_update_dict = self._make_epg_update_dict(name="updated-epg-name",
                                            add_tag=ptag["policy_tag"]["id"])
        epg_update_ret = plugin.update_endpoint_group(admin_context,
                                                      epg_ret["id"],
                                                      epg_update_dict)
        epg_update_dict["endpoint_group"]["id"] = epg_update_ret["id"]
        self.assertEqual(epg_update_dict["endpoint_group"], epg_update_ret)

    def test_create_update_security_group_with_policy_tag(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ptag = self._fake_policy_tag_dict()
        ptag_ret = plugin.create_policy_tag(admin_context, ptag)
        ptag["policy_tag"]["id"] = ptag_ret["id"]

        # Create security group
        sg = self._fake_sg()
        sg_ret = plugin.create_security_group(admin_context, sg)

        sg_update_dict = self._make_sg_update_dict(
                                             add_tag=ptag["policy_tag"]["id"])

        sg_update_ret = plugin.update_endpoint_group(admin_context,
                                                     sg_ret["id"],
                                                     sg_update_dict)

        sg_update_dict["endpoint_group"]["id"] = sg_update_ret["id"]
        (sg_update_dict["endpoint_group"]
         ["description"]) = sg_update_ret["description"]
        sg_update_dict["endpoint_group"]["name"] = sg_update_ret["name"]
        (sg_update_dict["endpoint_group"]
         ["is_security_group"]) = sg_update_ret["is_security_group"]
        (sg_update_dict["endpoint_group"]
         ["policy_tag_id"]) = sg_update_ret["policy_tag_id"]
        (sg_update_dict["endpoint_group"]
         ["remove_tag"]) = sg_update_ret["remove_tag"]
        self.assertEqual(sg_update_dict["endpoint_group"], sg_update_ret)

    def test_list_all_endpoint_groups(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        # Create first endpoint group
        epg1 = self._make_epg_dict()
        epg_ret1 = plugin.create_endpoint_group(
                      admin_context, epg1)
        epg1["endpoint_group"]["id"] = epg_ret1["id"]

        # Create second endpoint group
        ptag = self._fake_policy_tag_dict()
        ptag_ret = plugin.create_policy_tag(admin_context, ptag)
        ptag["policy_tag"]["id"] = ptag_ret["id"]

        epg2 = self._make_epg_dict(ptag_id=ptag_ret["id"])

        epg_ret2 = plugin.create_endpoint_group(
                      admin_context, epg2)
        epg2["endpoint_group"]["id"] = epg_ret2["id"]

        # List both endpoint groups
        epg_list_ret = plugin.get_endpoint_groups(admin_context)
        self.assertItemsEqual(epg_list_ret, [epg_ret1, epg_ret2])

    def _make_epg_dict(self, ptag_id=None,
                       name="test_sg_name"):

        epg_dict = {"endpoint_group": {
                   "tenant_id": "test_tenant",
                   "name": name,
                   "description": "test_description",
                   "is_security_group": False}}
        if ptag_id is not None:
            epg_dict["endpoint_group"]["policy_tag_id"] = ptag_id
        return epg_dict

    def _make_epg_update_dict(self, name="test_tenant",
                              description="test_description",
                              add_tag=[],
                              remove_tag=[]):
        return {"endpoint_group": {
                   "tenant_id": "test_tenant",
                   "name": name,
                   "description": description,
                   "add_tag": add_tag,
                   "remove_tag": remove_tag,
                   "is_security_group": False,
                   "policy_tag_id": None}}

    def _make_sg_update_dict(self, add_tag=[],
                             remove_tag=[]):
        sg_dict = {"endpoint_group": {
                   "tenant_id": "test_tenant"}}
        if len(add_tag) != 0:
            sg_dict["endpoint_group"]["add_tag"] = add_tag
        if len(remove_tag) != 0:
            sg_dict["endpoint_group"]["remove_tag"] = remove_tag
        return sg_dict

    def _fake_policy_tag_dict(self):
        return {"policy_tag": {
                               "tenant_id": "test_tenant",
                               "name": "test_name",
                               "tag_type": "dot1q",
                               "tag_id": "10-20",
                               "floatingip_id": None,
                               "floating_ip_address": None,
                               "router_id": None}}

    def _fake_sg(self):
        return {"security_group": {"name": "fake-sg",
                                   "description": "test_description",
                                   "tenant_id": "test_tenant"}}



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

from oslo_log import log as logging
from oslo_utils import excutils

from networking_plumgrid._i18n import _LI, _LW
from networking_plumgrid.neutron.plugins.common import exceptions as exception
from networking_plumgrid.neutron.plugins.common.locking import lock_object
from retrying import retry

LOG = logging.getLogger(__name__)

GL = "pg-gl"


def retry_if_res_error(excp):
    return isinstance(excp, exception.TenantResourcesInUse)


class PGLock(object):
    def __init__(self, context, uuid, ds=True):
        self.context = context
        self.uuid = uuid
        self.listener = None
        self.ds = ds

    @staticmethod
    def get_lock_id(self):
        return lock_object.PGLock.get_lock_id(self.uuid)

    def try_acquire(self):
        """
        Try to acquire a lock, but don't raise resource in use
        exception or try to steal lock.
        """
        return lock_object.PGLock.create(self.uuid)

    @retry(wait_random_min=500, wait_random_max=4000, stop_max_attempt_number=120, retry_on_exception=retry_if_res_error)  # noqa
    def acquire(self, retry=True):
        """
        Acquire a lock on the resource.

        :param retry: When True, retry if lock was released while stealing.
        :type retry: boolean
        """
        if not self.ds:
            return
        err_msg = ("Tenant (" + self.uuid + ") resources are currently in "
                   " use by another session. Please re-try")
        lock_id = lock_object.PGLock.create(self.uuid)
        if lock_id is None:
            LOG.debug("Lock acquired on resource "
                      "%(resource)s" % {'resource': self.uuid})
            return

        if (lock_id == self.uuid):
            expired_locks = False
            # Check for a potential expired locks and release them
            exp_locks = lock_object.PGLock.get(self.uuid)
            for l in exp_locks:
                expired_locks = True
                LOG.info(_LI("Releasing an expired lock for resource "
                         " %(resource)s"), {'resource': l.uuid})
                self.release(l.uuid)
            if not expired_locks:
                raise exception.TenantResourcesInUse(err_msg=err_msg)
            else:
                lock_id = lock_object.PGLock.create(self.uuid)
                if lock_id is None:
                    LOG.debug("Lock acquired on resource "
                              "%(resource)s" % {'resource': self.uuid})
                    return
                raise exception.TenantResourcesInUse(err_msg=err_msg)

    def release(self, uuid):
        """Release a lock."""
        # Only the resource that owns the lock will be releasing it.
        result = lock_object.PGLock.release(uuid)
        if result is True:
            LOG.warn(_LW("Lock was already released on resource %s!"), uuid)
        else:
            LOG.debug("Resource %(resource)s released "
                      "lock" % {'resource': self.uuid})

    @contextlib.contextmanager
    def thread_lock(self, uuid):
        """
        Acquire a lock and release it only if there is an exception.  The
        release method still needs to be scheduled to be run at the
        end of the thread using the Thread.link method.
        """
        try:
            self.acquire()
            yield
        except exception.TenantResourcesInUse:
            raise
        except:  # noqa
            with excutils.save_and_reraise_exception():
                LOG.debug("Locking contest seen in the catch all exc")

    @contextlib.contextmanager
    def try_thread_lock(self, uuid):
        """
        Similar to thread_lock, but acquire the lock using try_acquire
        and only release it upon any exception after a successful
        acquisition.
        """
        result = None
        try:
            result = self.try_acquire()
            yield result
        except:  # noqa
            if result is None:  # Lock was successfully acquired
                with excutils.save_and_reraise_exception():
                    self.release(uuid)
            raise



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from networking_plumgrid.neutron.plugins.common import \
    policy_exceptions as p_excep
from networking_plumgrid.neutron.plugins.db.policy.policy_tag_db \
    import PolicyTag
from networking_plumgrid.neutron.plugins.extensions import \
    endpointgroup as epgroup
from neutron.api.v2 import attributes
from neutron.db import common_db_mixin
from neutron.db import model_base
from neutron.db import models_v2
from neutron.db.securitygroups_db import SecurityGroup
from oslo_log import log as logging
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import exc

LOG = logging.getLogger(__name__)


class EndpointGroup(model_base.BASEV2, models_v2.HasId,
                    models_v2.HasTenant):
    """DB definition for PLUMgrid endpoint group object"""

    __tablename__ = "pg_endpoint_groups"

    name = sa.Column(sa.String(attributes.NAME_MAX_LEN))
    description = sa.Column(sa.String(255))
    policy_tag_id = sa.Column(sa.String(36),
                              sa.ForeignKey("pg_policy_tags.id",
                                            ondelete="RESTRICT"),
                              nullable=True)
    ptags = orm.relationship(PolicyTag,
                             backref=orm.backref('p_tag'),
                    primaryjoin="PolicyTag.id==EndpointGroup.policy_tag_id")


class SecurityPolicyTagBinding(model_base.BASEV2):
    """
    DB definition for storing Security Groups and
    Policy Tag mapping
    """
    __tablename__ = "pg_security_policy_tag_binding"

    security_group_id = sa.Column(sa.String(length=36),
                                  sa.ForeignKey("securitygroups.id",
                                                ondelete='CASCADE'),
                                  primary_key=True)
    policy_tag_id = sa.Column(sa.String(length=36),
                              sa.ForeignKey("pg_policy_tags.id",
                                            ondelete='RESTRICT'),
                              nullable=False)

    ep = orm.relationship(SecurityGroup,
              backref=orm.backref("sg_binding",
                                  lazy="joined",
                                  cascade="all,delete"),
  primaryjoin="SecurityGroup.id==SecurityPolicyTagBinding.security_group_id")


class EndpointGroupMixin(common_db_mixin.CommonDbMixin):
    def create_endpoint_group(self, context, endpoint_group):
        """
        Creates a endpoint group with with optional policy tag
        Args:
             endpoint_group:
                   JSON object with policy group attributes
                   name : display name policy tag
                   tenant_id : tenant uuid
                   id : endpoint group uuid
                   description : description for endpoint group
                   tag : policy tag uuid/name
        """
        ep_group = endpoint_group["endpoint_group"]

        if "policy_tag_id" in ep_group and ep_group["policy_tag_id"]:
            # Check policy tag exists
            ptag = self._check_policy_tag(context, ep_group["policy_tag_id"])

            # Check policy tag is not already in use
            if ptag["tag_type"].lower() == "fip":
                self._check_ptag_in_use(context, ep_group["policy_tag_id"])
        else:
            ep_group["policy_tag_id"] = None

        with context.session.begin(subtransactions=True):
            ep_group_db = EndpointGroup(tenant_id=ep_group["tenant_id"],
                                name=ep_group["name"],
                                policy_tag_id=ep_group["policy_tag_id"],
                                description=ep_group["description"])
            context.session.add(ep_group_db)
        return self._make_ep_grp_dict(ep_group_db)

    def get_endpoint_group(self, context, ep_grp_id, fields=None):
        """
        Gets an existing endpoint group
        Args:
             ep_grp_id = uuid of the policy group being requested
        """
        try:
            query = self._model_query(context, EndpointGroup)
            ep_grp_db = query.filter_by(id=ep_grp_id).one()
        except exc.NoResultFound:
            try:
                query = self._model_query(context,
                                          SecurityGroup)
                ep_grp_db = query.filter_by(id=ep_grp_id).one()
            except exc.NoResultFound:
                raise epgroup.NoEndpointGroupFound(id=ep_grp_id)
            return self._make_ep_grp_dict_from_sec_grp_obj(context, ep_grp_db)

        return self._make_ep_grp_dict(ep_grp_db, fields)

    def get_endpoint_groups(self, context, filters=None, fields=None,
             sorts=None, limit=None, marker=None, page_reverse=None):
        """
        Gets the list of all the existing endpoint groups
        """
        endpoint_groups = self._get_collection(context, EndpointGroup,
                                    self._make_ep_grp_dict, filters=filters,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker, fields=fields,
                                    page_reverse=page_reverse)
        security_groups = self.get_security_groups(context, filters=filters,
                                                   fields=fields)
        return self._make_security_group_to_epg_dict(context, security_groups,
                                                     endpoint_groups)

    def delete_endpoint_group(self, context, ep_grp_id):
        """
        Deletes an existing endpoint group
        Args:
             ep_grp_id = uuid of the endpoint group being deleted
        """
        try:
            query = context.session.query(EndpointGroup)
            ep_grp_db = query.filter_by(id=ep_grp_id).first()
        except exc.NoResultFound:
            raise epgroup.NoEndpointGroupFound(id=ep_grp_id)
        with context.session.begin(subtransactions=True):
            context.session.delete(ep_grp_db)

    def update_endpoint_group(self, context, epg_id, endpoint_group):
        """
        Updates an existing endpoint group
        Args:
             epg_id:
                   uuid of the endpoint group being updated
             endpoint_group:
                   JSON with updated attributes of the endpoint group
                   name : name of endpoint group
                   tenant_id : tenant uuid
                   id : uuid of endpoint group
        """
        epg = endpoint_group["endpoint_group"]
        is_security_group = False
        if not epg:
            raise epgroup.UpdateParametersRequired()

        if "add_tag" in epg and len(epg["add_tag"]) != 0:
            # Check policy tag exists
            ptag = self._check_policy_tag(context, epg["add_tag"])

            # Check policy tag is not already in use
            if ptag["tag_type"].lower() == "fip":
                self._check_ptag_in_use(context, epg["add_tag"])

        with context.session.begin(subtransactions=True):
            try:
                query = self._model_query(context, EndpointGroup)
                epg_db = query.filter_by(id=epg_id).one()
            except exc.NoResultFound:
                is_security_group = self._is_security_group(context, epg_id)
                if is_security_group is True:
                    query = self._model_query(context,
                                              SecurityGroup)
                    ep_grp_db = query.filter_by(id=epg_id).one()
                else:
                    raise epgroup.NoEndpointGroupFound(id=epg_id)
            if is_security_group is True:
                if 'name' in epg or 'description' in epg:
                    raise epgroup.SGUpdateDisallowed()
                if 'add_tag' in epg:
                    sg_grp_db = SecurityPolicyTagBinding(
                                            security_group_id=epg_id,
                                            policy_tag_id=epg["add_tag"])
                    context.session.add(sg_grp_db)
                if 'remove_tag' in epg:
                    try:
                        query = self._model_query(context,
                                                  SecurityPolicyTagBinding)
                        query.filter_by(policy_tag_id=epg["remove_tag"]).one()
                    except exc.NoResultFound:
                        raise p_excep.NoPolicyTagAssociation(
                                                         id=epg['remove_tag'],
                                                         epg_id=epg_id)
                    query = self._model_query(context,
                                              SecurityPolicyTagBinding)
                    query.filter_by(security_group_id=epg_id,
                                    policy_tag_id=epg["remove_tag"]).delete()
            else:
                if 'name' in epg:
                    epg_db.update({"name": epg["name"]})
                if 'description' in epg:
                    epg_db.update({"description": epg["description"]})
                if 'add_tag' in epg:
                    epg_db.update({"policy_tag_id": epg["add_tag"]})
                if 'remove_tag' in epg:
                    if epg_db["policy_tag_id"]:
                        epg_db["policy_tag_id"] = None
                    else:
                        raise p_excep.NoPolicyTagAssociation(
                                                         id=epg['remove_tag'],
                                                         epg_id=epg_id)
        if is_security_group is True:
            return self._make_ep_grp_dict_from_sec_grp_obj(context, ep_grp_db)
        else:
            return self._make_ep_grp_dict(epg_db)

    def _make_ep_grp_dict(self, epg, fields=None):
        ep_grp_dict = {"id": epg.id,
                       "name": epg.name,
                       "policy_tag_id": epg.policy_tag_id,
                       "description": epg.description,
                       "tenant_id": epg.tenant_id,
                       "is_security_group": False}
        return self._fields(ep_grp_dict, fields)

    def _make_ep_grp_dict_from_sec_grp_obj(self, context, sec_grp_obj):
        sg_map = self._get_security_policy_tag_binding(context,
                                                       sec_grp_obj["id"])
        epg_dict = {"id": sec_grp_obj["id"],
                    "name": sec_grp_obj["name"],
                    "description": sec_grp_obj["description"],
                    "policy_tag_id": None,
                    "tenant_id": sec_grp_obj["tenant_id"],
                    "is_security_group": True}
        if sg_map and sg_map["security_group_id"] == sec_grp_obj["id"]:
            epg_dict["policy_tag_id"] = sg_map["policy_tag_id"]
        return epg_dict

    def _make_security_group_to_epg_dict(self, context, secgrp_collection,
                                         epg_collection):
        """
        Make a dict to show security groups as part of
        endpoint group list
        Args:
            secgrp_collection: List of security groups
        """
        for sec_grp in secgrp_collection:
            if "name" not in sec_grp:
                sec_grp = self._make_security_group_dict(
                              self._get_security_group(context, sec_grp['id']))
            sg_map = self._get_security_policy_tag_binding(context,
                                                           sec_grp["id"])
            epg_dict = {"id": sec_grp["id"],
                        "name": sec_grp["name"],
                        "description": sec_grp["description"],
                        "policy_tag_id": None,
                        "tenant_id": sec_grp["tenant_id"],
                        "is_security_group": True}
            if sg_map and sg_map["security_group_id"] == sec_grp["id"]:
                epg_dict["policy_tag_id"] = sg_map["policy_tag_id"]
            epg_collection.append(epg_dict)
        return epg_collection

    def _check_policy_tag(self, context, ptag_id):
        query = context.session.query(PolicyTag)
        try:
            ptag = query.filter_by(id=ptag_id).one()
        except exc.NoResultFound:
            raise p_excep.NoPolicyTagFound(id=ptag_id)
        return ptag

    def _check_ptag_in_use(self, context, ptag_id):
        query = context.session.query(EndpointGroup)
        try:
            ptag = query.filter_by(policy_tag_id=ptag_id).one()
            if ptag:
                raise p_excep.PolicyTagAlreadyInUse(ptag_id=ptag_id,
                                                    epg_id=ptag.id)
        except exc.MultipleResultsFound:
            ptag = query.filter_by(policy_tag_id=ptag_id)
            if ptag:
                raise p_excep.PolicyTagAlreadyInUseMultipleEPG(ptag_id=ptag_id)
        except exc.NoResultFound:
            try:
                query = self._model_query(context,
                                          SecurityPolicyTagBinding)
                sg_map = query.filter_by(policy_tag_id=ptag_id).one()
                if sg_map:
                    raise p_excep.PolicyTagAlreadyInUseSG(ptag_id=ptag_id,
                                              sg_id=sg_map.security_group_id)
            except exc.MultipleResultsFound:
                sg_map = query.filter_by(policy_tag_id=ptag_id)
                if sg_map:
                    raise p_excep.PolicyTagAlreadyInUseMultipleSG(
                                                              ptag_id=ptag_id)
            except exc.NoResultFound:
                pass

    def _is_security_group(self, context, epg_id):
        epg_id_list = self.get_endpoint_groups(context,
                                               filters={'id': [epg_id]},
                                               fields=["id"])
        if len(epg_id_list) == 1:
            if ("is_security_group" in epg_id_list[0]
                and epg_id_list[0]["is_security_group"]):
                return True

    def _get_security_policy_tag_binding(self, context, sg_id):
        query = self._model_query(context,
                                  SecurityPolicyTagBinding)
        sg_map = {}
        try:
            sg_map = query.filter_by(security_group_id=sg_id).one()
            return sg_map
        except exc.NoResultFound:
            return sg_map



# Copyright 2016 PLUMgrid, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Policy Service extension unit tests
"""
from networking_plumgrid.neutron.plugins.common import \
    policy_exceptions as policy_excep
from networking_plumgrid.neutron.plugins.extensions import \
    policyservice as ext_ps
from networking_plumgrid.neutron.tests.unit import \
    test_networking_plumgrid as test_pg

from neutron.api import extensions
from neutron.api.v2 import attributes
from neutron.common import exceptions as n_excep
from neutron import context
from neutron import manager
from neutron.tests.unit.api import test_extensions as test_ext
from oslo_log import log as logging
import uuid

LOG = logging.getLogger(__name__)


class PolicyServiceExtensionManager(object):

    def get_resources(self):
        return ext_ps.Policyservice.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class PolicyServiceTestCase(test_pg.PLUMgridPluginV2TestCase):
    def setUp(self, plugin=None, ext_mgr=None):
        super(PolicyServiceTestCase, self).setUp()
        ext_mgr = PolicyServiceExtensionManager()
        extensions.PluginAwareExtensionManager._instance = None
        self.ext_api = test_ext.setup_extensions_middleware(ext_mgr)


class TestPolicyService(PolicyServiceTestCase):

    def test_create_policy_service_name(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ps = self._make_ps_dict()

        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps["policy_service"]["id"] = ps_ret["id"]
        self.assertEqual(ps_ret, ps["policy_service"])

    def test_create_policy_service_inport(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(inports=[{'id': port_ret['id']}])

        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps["policy_service"]["id"] = ps_ret["id"]
        self.assertEqual(ps_ret, ps["policy_service"])

    def test_create_policy_service_eport(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(eports=[{'id': port_ret['id']}])

        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps["policy_service"]["id"] = ps_ret["id"]
        self.assertEqual(ps_ret, ps["policy_service"])

    def test_create_policy_service_bidirect_port(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(bidirect_ports=[{'id': port_ret['id']}])
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps["policy_service"]["id"] = ps_ret["id"]
        self.assertEqual(ps_ret, ps["policy_service"])

    def test_create_policy_service_inport_name(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(inports=[{'id': port_ret['name']}])

        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps["policy_service"]["id"] = ps_ret["id"]
        self.assertEqual(ps_ret, ps["policy_service"])

    def test_create_policy_service_eport_name(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(eports=[{'id': port_ret['name']}])

        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps["policy_service"]["id"] = ps_ret["id"]
        self.assertEqual(ps_ret, ps["policy_service"])

    def test_create_policy_service_bidirect_port_name(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(bidirect_ports=[{'id': port_ret['name']}])
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps["policy_service"]["id"] = ps_ret["id"]
        self.assertEqual(ps_ret, ps["policy_service"])

    def test_create_policy_service_dup_ports_inport(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(inports=[{'id': port_ret['id']},
                                         {'id': port_ret['id']}])
        self.assertRaises(policy_excep.DuplicatePortFound,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_dup_ports_eport(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(eports=[{'id': port_ret['id']},
                                {'id': port_ret['id']}])
        self.assertRaises(policy_excep.DuplicatePortFound,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_dup_ports_bidirect(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(bidirect_ports=[{'id': port_ret['id']},
                                                {'id': port_ret['id']}])
        self.assertRaises(policy_excep.DuplicatePortFound,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_dup_ports_across_config(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(eports=[{'id': port_ret['id']}],
                                inports=[{'id': port_ret['id']}])
        self.assertRaises(policy_excep.DuplicatePortFound,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_without_leg_mode(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ps = self._make_ps_dict()

        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps["policy_service"]["id"] = ps_ret["id"]
        self.assertEqual(ps_ret, ps["policy_service"])

    def test_create_policy_service_inport_owner_not_nova(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"], device_owner="invalid:owner")
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(inports=[{'id': port_ret['id']}])
        self.assertRaises(policy_excep.InvalidPortForPolicyService,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_eport_owner_not_nova(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"], device_owner="invalid:owner")
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(eports=[{'id': port_ret['id']}])
        self.assertRaises(policy_excep.InvalidPortForPolicyService,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_bidirect_port_owner_not_nova(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"], device_owner="invalid:owner")
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(bidirect_ports=[{'id': port_ret['id']}])
        self.assertRaises(policy_excep.InvalidPortForPolicyService,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_inport_port_not_found(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        ps = self._make_ps_dict(inports=[{'id': str(uuid.uuid4())}])
        self.assertRaises(n_excep.PortNotFound,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_eport_port_not_found(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        ps = self._make_ps_dict(eports=[{'id': str(uuid.uuid4())}])
        self.assertRaises(n_excep.PortNotFound,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_bidirect_port_not_found(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        ps = self._make_ps_dict(bidirect_ports=[{'id': str(uuid.uuid4())}])
        self.assertRaises(n_excep.PortNotFound,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_port_already_in_use_ps(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps1 = self._make_ps_dict(inports=[{'id': port_ret['id']}])

        plugin.create_policy_service(admin_context, ps1)
        ps2 = self._make_ps_dict(name="test_ps2",
                                 inports=[{'id': port_ret['id']}])
        self.assertRaises(policy_excep.PortAlreadyInUsePolicyService,
                          plugin.create_policy_service, admin_context, ps2)

    def test_create_policy_service_inport_bidirect_port(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port1 = self._fake_port(network_ret["id"])
        port_ret1 = plugin.create_port(
                      admin_context, port1)
        port2 = self._fake_port(network_ret["id"])
        port_ret2 = plugin.create_port(
                      admin_context, port2)
        ps = self._make_ps_dict(inports=[{'id': port_ret1['id']}],
                                bidirect_ports=[{'id': port_ret2['id']}])
        self.assertRaises(policy_excep.InvalidPortConfigPolicyService,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_eport_bidirect_port(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port1 = self._fake_port(network_ret["id"])
        port_ret1 = plugin.create_port(
                      admin_context, port1)
        port2 = self._fake_port(network_ret["id"])
        port_ret2 = plugin.create_port(
                      admin_context, port2)
        ps = self._make_ps_dict(eports=[{'id': port_ret1['id']}],
                                bidirect_ports=[{'id': port_ret2['id']}])
        self.assertRaises(policy_excep.InvalidPortConfigPolicyService,
                          plugin.create_policy_service, admin_context, ps)

    def test_create_policy_service_inport_eport_bidirect_port(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port1 = self._fake_port(network_ret["id"])
        port_ret1 = plugin.create_port(
                      admin_context, port1)
        port2 = self._fake_port(network_ret["id"])
        port_ret2 = plugin.create_port(
                      admin_context, port2)
        port3 = self._fake_port(network_ret["id"])
        port_ret3 = plugin.create_port(
                      admin_context, port3)
        ps = self._make_ps_dict(eports=[{'id': port_ret1['id']}],
                                inports=[{'id': port_ret2['id']}],
                                bidirect_ports=[{'id': port_ret3['id']}])
        self.assertRaises(policy_excep.InvalidPortConfigPolicyService,
                          plugin.create_policy_service, admin_context, ps)

    def test_delete_policy_service(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ps = self._make_ps_dict()

        ps_ret = plugin.create_policy_service(admin_context, ps)
        resp = plugin.delete_policy_service(admin_context, ps_ret["id"])
        self.assertEqual(None, resp)

    def test_delete_policy_service_ports_in_use(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(inports=[{'id': port_ret['id']}])

        ps_ret = plugin.create_policy_service(admin_context, ps)
        self.assertRaises(policy_excep.PolicyServiceInUse,
                          plugin.delete_policy_service, admin_context,
                          ps_ret["id"])

    def test_delete_policy_service_does_not_exist(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        self.assertRaises(policy_excep.NoPolicyServiceFound,
                          plugin.delete_policy_service, admin_context,
                          str(uuid.uuid4()))

    def test_update_policy_service_name(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ps = self._make_ps_dict()
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict()
        ps_update_ret = plugin.update_policy_service(admin_context,
                                                     ps_ret["id"],
                                                     ps_update_dict)
        self.assertEqual(ps_update_dict["policy_service"]["name"],
                         ps_update_ret["name"])

    def test_update_policy_service_inport(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict()
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_inports=
                                                   [{"id": port_ret["id"]}])
        ps_update_ret = plugin.update_policy_service(admin_context,
                                                     ps_ret["id"],
                                                     ps_update_dict)
        self.assertEqual(ps_update_dict["policy_service"]["add_ingress_ports"],
                         ps_update_ret["ingress_ports"])

    def test_update_policy_service_eport(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict()
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_eports=
                                                   [{"id": port_ret["id"]}])
        ps_update_ret = plugin.update_policy_service(admin_context,
                                                     ps_ret["id"],
                                                     ps_update_dict)
        self.assertEqual(ps_update_dict["policy_service"]["add_egress_ports"],
                         ps_update_ret["egress_ports"])

    def test_update_policy_service_bidirect_port(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict()
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_bidirect_ports=
                                                   [{"id": port_ret["id"]}])
        ps_update_ret = plugin.update_policy_service(admin_context,
                                                     ps_ret["id"],
                                                     ps_update_dict)
        self.assertEqual(ps_update_dict["policy_service"]
                         ["add_bidirectional_ports"],
                         ps_update_ret["bidirectional_ports"])

    def test_update_policy_service_add_dup_inports(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict()
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_inports=
                                                   [{"id": port_ret["id"]},
                                                    {"id": port_ret["id"]}])
        self.assertRaises(policy_excep.DuplicatePortFound,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_add_dup_eports(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict()
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_eports=
                                                   [{"id": port_ret["id"]},
                                                   {"id": port_ret["id"]}])
        self.assertRaises(policy_excep.DuplicatePortFound,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_add_dup_bidirect_ports(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict()
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_bidirect_ports=
                                                   [{"id": port_ret["id"]},
                                                   {"id": port_ret["id"]}])
        self.assertRaises(policy_excep.DuplicatePortFound,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_remove_dup_inports(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(inports=[{'id': port_ret['id']}])
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(remove_inports=
                                                   [{"id": port_ret["id"]},
                                                   {"id": port_ret["id"]}])
        self.assertRaises(policy_excep.DuplicatePortFound,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_remove_dup_eports(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(eports=[{'id': port_ret['id']}])
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(remove_eports=
                                                   [{"id": port_ret["id"]},
                                                   {"id": port_ret["id"]}])
        self.assertRaises(policy_excep.DuplicatePortFound,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_remove_dup_bidirect_ports(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(bidirect_ports=[{'id': port_ret['id']}])
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(remove_bidirect_ports=
                                                   [{"id": port_ret["id"]},
                                                   {"id": port_ret["id"]}])
        self.assertRaises(policy_excep.DuplicatePortFound,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_dup_ports(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(inports=[{'id': port_ret['id']}])
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(remove_inports=
                                                   [{"id": port_ret["id"]}],
                                                   remove_eports=
                                                   [{"id": port_ret["id"]}])
        self.assertRaises(policy_excep.DuplicatePortFound,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_add_inports_owner_not_nova(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"], device_owner="invalid:owner")
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict()
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_inports=
                                                   [{"id": port_ret["id"]}])
        self.assertRaises(policy_excep.InvalidPortForPolicyService,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_add_eports_owner_not_nova(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"], device_owner="invalid:owner")
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict()
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_eports=
                                                   [{"id": port_ret["id"]}])
        self.assertRaises(policy_excep.InvalidPortForPolicyService,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_add_bidirect_ports_owner_not_nova(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"], device_owner="invalid:owner")
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict()
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_bidirect_ports=
                                                   [{"id": port_ret["id"]}])
        self.assertRaises(policy_excep.InvalidPortForPolicyService,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_add_inport_already_in_use(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(inports=[{'id': port_ret['id']}])

        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_inports=
                                                   [{"id": port_ret["id"]}])
        self.assertRaises(policy_excep.PortAlreadyInUsePolicyService,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_add_eport_already_in_use(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(inports=[{'id': port_ret['id']}])

        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_eports=
                                                   [{"id": port_ret["id"]}])
        self.assertRaises(policy_excep.PortAlreadyInUsePolicyService,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_update_policy_service_add_bport_already_in_use(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        port = self._fake_port(network_ret["id"])
        port_ret = plugin.create_port(
                      admin_context, port)
        ps = self._make_ps_dict(bidirect_ports=[{'id': port_ret['id']}])

        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_update_dict = self._make_ps_update_dict(add_bidirect_ports=
                                                   [{"id": port_ret["id"]}])
        self.assertRaises(policy_excep.PortAlreadyInUsePolicyService,
                          plugin.update_policy_service, admin_context,
                          ps_ret["id"], ps_update_dict)

    def test_create_get_policy_service(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        ps = self._make_ps_dict()
        ps_ret = plugin.create_policy_service(admin_context, ps)
        ps_get_ret = plugin.get_policy_service(admin_context, ps_ret["id"])
        self.assertEqual(ps_ret, ps_get_ret)

    def test_list_policy_services(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ps1 = self._make_ps_dict(name="list_ps1")
        ps_ret1 = plugin.create_policy_service(admin_context, ps1)
        ps2 = self._make_ps_dict(name="list_ps2")
        ps_ret2 = plugin.create_policy_service(admin_context, ps2)
        ps3 = self._make_ps_dict(name="list_ps3")
        ps_ret3 = plugin.create_policy_service(admin_context, ps3)

        ps_list_ret = plugin.get_policy_services(admin_context)
        self.assertItemsEqual(ps_list_ret, [ps_ret1, ps_ret2, ps_ret3])

    def _make_ps_dict(self, name="test_ps", inports=[], eports=[],
                      bidirect_ports=[]):
        return {"policy_service": {
                    "tenant_id": "test_tenant",
                    "description": "test-ps",
                    "name": name,
                    "ingress_ports": inports,
                    "egress_ports": eports,
                    "bidirectional_ports": bidirect_ports}}

    def _make_ps_update_dict(self, name="updated_ps", add_inports=[],
                             add_eports=[], add_bidirect_ports=[],
                             remove_inports=[], remove_eports=[],
                             remove_bidirect_ports=[]):
        update_dict = {"policy_service": {
                         "tenant_id": "test_tenant",
                         "name": name}}
        if add_inports:
            update_dict["policy_service"]["add_ingress_ports"] = add_inports
        if add_eports:
            update_dict["policy_service"]["add_egress_ports"] = add_eports
        if add_bidirect_ports:
            update_dict["policy_service"]["add_bidirectional_ports"] = \
                add_bidirect_ports
        if remove_inports:
            update_dict["policy_service"]["remove_ingress_ports"] = \
                remove_inports
        if remove_eports:
            update_dict["policy_service"]["remove_egress_ports"] = \
                remove_eports
        if remove_bidirect_ports:
            update_dict["policy_service"]["remove_bidirectional_ports"] = \
                remove_bidirect_ports
        return update_dict

    def _fake_network(self):
        data = {'network': {'name': "fake_network",
                            'admin_state_up': False,
                            'shared': False,
                            'router:external': [],
                            'provider:network_type': None,
                            'provider:segmentation_id': None,
                            'provider:physical_network': None}}
        return data

    def _fake_subnet(self, net_id):
        allocation_pools = [{'start': '10.0.0.2',
                             'end': '10.0.0.254'}]
        return {'subnet': {'name': "fake_subnet",
                           'network_id': net_id,
                           'gateway_ip': '10.0.0.1',
                           'dns_nameservers': ['10.0.0.2'],
                           'host_routes': [],
                           'cidr': '10.0.0.0/24',
                           'allocation_pools': allocation_pools,
                           'enable_dhcp': True,
                           'ip_version': 4}}

    def _fake_port(self, net_id, name="fake_port", security_groups = [],
                   device_owner="compute:nova"):
        return {'port': {'name': name,
                         'network_id': net_id,
                         'mac_address': attributes.ATTR_NOT_SPECIFIED,
                         'fixed_ips': attributes.ATTR_NOT_SPECIFIED,
                         'admin_state_up': True,
                         'device_id': 'fake_device_id',
                         'device_owner': device_owner,
                         'tenant_id': "fake_tenant",
                         'security_group': security_groups}}

    def assertEqualUpdate(self, ps_update_dict, ps_update_ret):
        def make_ps_dict(ps_db):
            if "add_ingress_ports" in ps_db:
                del ps_db["add_ingress_ports"]
            if "add_egress_ports" in ps_db:
                del ps_db["add_egress_ports"]
            if "add_bidirectional_ports" in ps_db:
                del ps_db["add_bidirectional_ports"]
            if "remove_ingress_ports" in ps_db:
                del ps_db["remove_ingress_ports"]
            if "remove_egress_ports" in ps_db:
                del ps_db["remove_egress_ports"]
            if "remove_bidirectional_ports" in ps_db:
                del ps_db["remove_bidirectional_ports"]
            return ps_db
        ps_ref = make_ps_dict(ps_update_dict["policy_service"])
        ps_actual = make_ps_dict(ps_update_ret)
        ps_ref["id"] = ps_actual["id"]
        ps_ref["name"] = ps_actual["name"]
        ps_ref["description"] = ps_actual["description"]
        ps_ref["ingress_ports"] = ps_actual["ingress_ports"]
        ps_ref["egress_ports"] = ps_actual["egress_ports"]
        ps_ref["bidirectional_ports"] = ps_actual["bidirectional_ports"]
        return self.assertEqual(ps_ref, ps_actual)



# Copyright 2015 PLUMgrid, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


SEGMENTATION_ID_DELIMITER = "#"
INTERFACE_SEG_ID_DELIMITER = "|"
DEVICE_INTERFACE_DELIMITER = "::"
DEVICE_DELIMITER = ","
INTERFACE_DELIMITER = ";"
"""
  Sample for providing input for gateway creation in config is noted below
  Options provide flexibility to user to create l2gateway
  For single device ,single interface with single vlan
    l2gw_switch = device_name1::int_name1|vlan1
  For single device multiple interfaces with single or multiple vlans
    l2gw_switch = device_name1::int_name1|vlan1#vlan2;int_name2|vlan3
  For multiple devices with mutiple interfaces having single or mutiple vlan
    l2gw_switch = device_name1::int_name1|vlan1,device_name2::int_name2|vlan2
"""


def get_interface(interfaces):
    interface_dict = []
    for interface in interfaces:
        if INTERFACE_SEG_ID_DELIMITER in interface:
            (int_name, segid) = interface.split(INTERFACE_SEG_ID_DELIMITER)
            if SEGMENTATION_ID_DELIMITER in segid:
                segid = segid.split(SEGMENTATION_ID_DELIMITER)
            else:
                segid = [segid]
            interface_detail = {'name': int_name, 'segmentation_id': segid}
        else:
            interface_detail = {'name': interface}
        interface_dict.append(interface_detail)
    return interface_dict


def get_l2gw_body(l2gw_conf):
    device_dict = []
    devices = l2gw_conf.split(DEVICE_DELIMITER)
    for device in devices:
        if DEVICE_INTERFACE_DELIMITER in device:
            (device_name, device_ip, interface) = device.split(
                                                   DEVICE_INTERFACE_DELIMITER)
            if INTERFACE_DELIMITER in interface:
                interface_dict = interface.split(INTERFACE_DELIMITER)
                interfaces = get_interface(interface_dict)
            else:
                interfaces = get_interface([interface])
        device = {'device_name': device_name,
                  'device_ip': device_ip,
                  'interfaces': interfaces}
        device_dict.append(device)
    body = {'devices': device_dict}
    return body


def form_dict_devices(devices):
    seg_ids = []
    device_dict = dict()
    int_seg = []
    for device in devices:
        device_name = device['device_name']
        interfaces = device['interfaces']
        for interface in interfaces:
            interface_name = interface['name']
            int_seg.append(interface_name)
            seg_id = interface['segmentation_id']
            if type(seg_id) is list:
                for segid in seg_id:
                    seg_ids.append(segid)
            else:
                seg_ids.append(seg_id)
            int_seg.append(seg_id)
            device_dict.setdefault(device_name, []).append(int_seg)
            int_seg = []
    return device_dict



# Copyright 2015 PLUMgrid, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import networking_plumgrid
from networking_plumgrid._i18n import _
from neutron.api import extensions
from neutron.api.v2 import attributes as attr
from neutron.api.v2 import base
from neutron.common import exceptions as nexceptions
from neutron import manager


PG_PAP = 'physical_attachment_point'
PG_PAPS = 'physical-attachment-points'

extensions.append_api_extensions_path(
    networking_plumgrid.neutron.plugins.extensions.__path__)


class InvalidInterfaceFormat(nexceptions.InvalidInput):
    message = _("Invalid format for interfaces: ")

    def __init__(self, error):
        super(InvalidInterfaceFormat, self).__init__()
        self.msg = self.message + error


class UpdateParametersRequired(nexceptions.InvalidInput):
    message = _("No update parameter specified atleast one needed")


class InvalidLacpValue(nexceptions.InvalidInput):
    message = _("LACP value must be set to True for hash_mode %(hash_mode)s")


class InvalidLacpValueAs(nexceptions.InvalidInput):
    message = _("LACP must be set to False to enable Active/Standby mode")


class InvalidAsValueLacp(nexceptions.InvalidInput):
    message = _("Active/Standby must be set to False to enable LACP mode")


class NoPhysicalAttachmentPointFound(nexceptions.NotFound):
    message = _("Physical Attachment Point with id %(id)s does not exist")


class TransitDomainLimit(nexceptions.InvalidInput):
    message = _("Only one physical attachment point"
                " allowed per transit domain.")


class InterfaceInUse(nexceptions.InvalidInput):
    message = _("Interface %(ifc)s is already in use by physical"
                " attachment point with id %(id)s")


def _validate_interfaces_list(data, valid_values=None):
    if not isinstance(data, list):
        raise InvalidInterfaceFormat("Must be a List of Interfaces")

    for interface in data:
        if 'hostname' not in interface:
            raise InvalidInterfaceFormat("hostname field is required")

        if 'interface' not in interface:
            raise InvalidInterfaceFormat("interface_name field is required")


attr.validators['type:validate_interfaces_list'] = _validate_interfaces_list


RESOURCE_ATTRIBUTE_MAP = {
    'physical_attachment_point': {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {
                'type:uuid': None
            },
            'is_visible': True,
            'primary_key': True
        },
        'name': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'default': ''
        },
        'interfaces': {
            'allow_post': True,
            'allow_put': False,
            'is_visible': True,
            'default': [],
            'validate': {
                'type:validate_interfaces_list': None
            }
        },
        'add_interfaces': {
            'allow_post': False,
            'allow_put': True,
            'is_visible': True,
            'default': [],
            'validate': {
                'type:validate_interfaces_list': None
            }
        },
        'remove_interfaces': {
            'allow_post': False,
            'allow_put': True,
            'is_visible': True,
            'default': [],
            'validate': {
                'type:validate_interfaces_list': None
            }
        },
        'hash_mode': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'default': 'L2',
            'validate': {
                'type:values': ['L2', 'L2+L3', 'L3', 'L3+L4']
            }
        },
        'lacp': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'default': False,
            'validate': {
                 'type:boolean': None
            }
        },
        'active_standby': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'default': False,
            'validate': {
                 'type:boolean': None
            }
        },
        'implicit': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': False,
            'default': False,
            'validate': {
                 'type:boolean': None
            }
        },
        'transit_domain_id': {
            'allow_post': True,
            'allow_put': False,
            'is_visible': True,
            'default': None,
        },
        'tenant_id': {
            'allow_post': True,
            'allow_put': False,
            'is_visible': True
        },
    }
}


class Physical_attachment_point(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Physical Attachment Point"

    @classmethod
    def get_alias(cls):
        return "physical-attachment-point"

    @classmethod
    def get_description(cls):
        return "This API will be used to configure physical attachment" \
               "points from Neutron and map it to Neutron external and" \
               "provider networks"

    @classmethod
    def get_namespace(cls):
        return "http://docs.openstack.org/ext/physical_attachment_points" \
               "/api/v2.0"

    @classmethod
    def get_updated(cls):
        return "2012-10-05T10:00:00-00:00"

    @classmethod
    def get_resources(cls):
        exts = []
        plugin = manager.NeutronManager.get_plugin()

        resource_name = PG_PAP
        collection_name = PG_PAPS
        params = RESOURCE_ATTRIBUTE_MAP.get(resource_name, dict())
        controller = base.create_resource(
            collection_name, resource_name, plugin, params)
        ex = extensions.ResourceExtension(collection_name,
                                          controller)
        exts.append(ex)
        return exts



# Copyright 2016 PLUMgrid, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Policy Tag extension unit tests
"""
from networking_plumgrid.neutron.plugins.common import \
    policy_exceptions as policy_excep
from networking_plumgrid.neutron.plugins.extensions import \
    policytag
from networking_plumgrid.neutron.tests.unit import \
    test_networking_plumgrid as test_pg

from neutron.api import extensions
from neutron import context
from neutron.extensions import l3
from neutron import manager
from neutron.tests.unit.api import test_extensions as test_ext
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class PolicyTagExtensionManager(object):

    def get_resources(self):
        return policytag.Policytag.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class PolicyTagTestCase(test_pg.PLUMgridPluginV2TestCase):
    def setUp(self, plugin=None, ext_mgr=None):
        super(PolicyTagTestCase, self).setUp()
        ext_mgr = PolicyTagExtensionManager()
        extensions.PluginAwareExtensionManager._instance = None
        self.ext_api = test_ext.setup_extensions_middleware(ext_mgr)


class TestPolicyTag(PolicyTagTestCase):
    def test_create_policy_tag_with_fip(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        subnet = self._fake_subnet(network_ret["id"])
        subnet["subnet"]["tenant_id"] = tenant_context.tenant_id
        plugin.create_subnet(tenant_context, subnet)

        floatingip = self._make_floatingip_dict(network_ret)
        floatingip_ret = plugin.create_floatingip(admin_context, floatingip)
        ptag = self._make_ptag_dict(tag_type="fip",
                    floatingip_id=floatingip_ret["id"],
                    floating_ip_address=floatingip_ret["floating_ip_address"],
                    router_id="testrouter")

        ptag_ret = plugin.create_policy_tag(
                      admin_context, ptag)
        ptag["policy_tag"]["id"] = ptag_ret["id"]
        self.assertEqual(ptag_ret, ptag["policy_tag"])

    def test_create_policy_tag_with_dot1q(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ptag = self._make_ptag_dict(tag_type="dot1q",
                                    tag_id="100-200")

        ptag_ret = plugin.create_policy_tag(
                      admin_context, ptag)
        ptag["policy_tag"]["id"] = ptag_ret["id"]
        self.assertEqual(ptag_ret, ptag["policy_tag"])

    def test_create_policy_tag_with_nsh(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ptag = self._make_ptag_dict(tag_type="nsh",
                                    tag_id="10")

        ptag_ret = plugin.create_policy_tag(
                      admin_context, ptag)
        ptag["policy_tag"]["id"] = ptag_ret["id"]
        self.assertEqual(ptag_ret, ptag["policy_tag"])

    def test_create_policy_tag_with_no_type(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ptag = self._make_ptag_dict(tag_type=None)

        self.assertRaises(policy_excep.InvalidPolicyTagType,
                          plugin.create_policy_tag, admin_context, ptag)

    def test_create_policy_tag_with_non_existing_fip(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ptag = self._make_ptag_dict(tag_type="fip",
                                    floatingip_id="testid",
                                    floating_ip_address="10.0.0.1",
                                    router_id="test-router")

        self.assertRaises(l3.FloatingIPNotFound,
                          plugin.create_policy_tag, admin_context, ptag)

    def test_create_policy_tag_in_use_fip(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        subnet = self._fake_subnet(network_ret["id"])
        subnet["subnet"]["tenant_id"] = tenant_context.tenant_id
        plugin.create_subnet(tenant_context, subnet)

        floatingip = self._make_floatingip_dict(network_ret)
        floatingip_ret = plugin.create_floatingip(admin_context, floatingip)
        ptag = self._make_ptag_dict(tag_type="fip",
                    floatingip_id=floatingip_ret["id"],
                    floating_ip_address=floatingip_ret["floating_ip_address"],
                    router_id="testrouter")

        # Create first policy tag with Floating IP
        plugin.create_policy_tag(
                      admin_context, ptag)

        # Create second policy tag with same Floating IP
        self.assertRaises(policy_excep.FloatingIPAlreadyInUse,
                          plugin.create_policy_tag, admin_context, ptag)

    def test_create_delete_policy_tag(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        tenant_context = context.Context('', 'not_admin')

        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        subnet = self._fake_subnet(network_ret["id"])
        subnet["subnet"]["tenant_id"] = tenant_context.tenant_id
        plugin.create_subnet(tenant_context, subnet)

        floatingip = self._make_floatingip_dict(network_ret)
        floatingip_ret = plugin.create_floatingip(admin_context, floatingip)
        ptag = self._make_ptag_dict(tag_type="fip",
                    floatingip_id=floatingip_ret["id"],
                    floating_ip_address=floatingip_ret["floating_ip_address"],
                    router_id="testrouter")

        ptag_ret = plugin.create_policy_tag(
                      admin_context, ptag)

        resp = plugin.delete_policy_tag(admin_context, ptag_ret["id"])
        self.assertEqual(None, resp)

    def test_create_update_policy_tag(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        ptag = self._make_ptag_dict(tag_type="dot1q",
                                    tag_id="100-200")

        ptag_ret = plugin.create_policy_tag(
                      admin_context, ptag)

        ptag_update_dict = self._make_ptag_update_dict(
                                                   name="updated-policy-tag")
        ptag_update_ret = plugin.update_policy_tag(admin_context,
                                                   ptag_ret["id"],
                                                   ptag_update_dict)
        ptag_update_dict["policy_tag"]["id"] = ptag_update_ret["id"]
        ptag_update_dict["policy_tag"]["tag_id"] = ptag_update_ret["tag_id"]
        (ptag_update_dict["policy_tag"]
         ["tag_type"]) = ptag_update_ret["tag_type"]
        (ptag_update_dict["policy_tag"]
         ["router_id"]) = ptag_update_ret["router_id"]
        (ptag_update_dict["policy_tag"]
         ["floatingip_id"]) = ptag_update_ret["floatingip_id"]
        (ptag_update_dict["policy_tag"]
         ["floating_ip_address"]) = ptag_update_ret["floating_ip_address"]
        self.assertEqual(ptag_update_dict["policy_tag"], ptag_update_ret)

    def test_list_policy_tags(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        tenant_context = context.Context('', 'not_admin')

        # Create policy Tag with Floating IP
        network = self._fake_network()
        network["network"]["tenant_id"] = tenant_context.tenant_id
        network_ret = plugin.create_network(tenant_context, network)

        subnet = self._fake_subnet(network_ret["id"])
        subnet["subnet"]["tenant_id"] = tenant_context.tenant_id
        plugin.create_subnet(tenant_context, subnet)

        floatingip = self._make_floatingip_dict(network_ret)
        floatingip_ret = plugin.create_floatingip(admin_context, floatingip)
        ptag1 = self._make_ptag_dict(tag_type="fip",
                    floatingip_id=floatingip_ret["id"],
                    floating_ip_address=floatingip_ret["floating_ip_address"],
                    router_id="testrouter")

        ptag_ret1 = plugin.create_policy_tag(
                      admin_context, ptag1)

        # Create Policy Tag with dot1q
        ptag2 = self._make_ptag_dict(tag_type="dot1q",
                                    tag_id="100-200")

        ptag_ret2 = plugin.create_policy_tag(
                      admin_context, ptag2)

        # Create Policy Tag with nsh
        ptag3 = self._make_ptag_dict(tag_type="nsh",
                                    tag_id="10")

        ptag_ret3 = plugin.create_policy_tag(
                      admin_context, ptag3)

        ptag_list_ret = plugin.get_policy_tags(admin_context)
        self.assertItemsEqual(ptag_list_ret, [ptag_ret1, ptag_ret2, ptag_ret3])

    def _make_ptag_dict(self, tag_type, tag_id=None,
                        floatingip_id=None, floating_ip_address=None,
                        router_id=None):
        return {"policy_tag": {
                               "tenant_id": "test_tenant",
                               "name": "test_name",
                               "tag_type": tag_type,
                               "tag_id": tag_id,
                               "floatingip_id": floatingip_id,
                               "floating_ip_address": floating_ip_address,
                               "router_id": router_id}}

    def _fake_network(self):
        data = {'network': {'name': "fake_network",
                            'admin_state_up': True,
                            'shared': False,
                            'router:external': True,
                            'provider:network_type': None,
                            'provider:segmentation_id': None,
                            'provider:physical_network': None}}
        return data

    def _fake_subnet(self, net_id):
        allocation_pools = [{'start': '10.0.0.2',
                             'end': '10.0.0.254'}]
        return {'subnet': {'name': "fake_subnet",
                           'network_id': net_id,
                           'gateway_ip': '10.0.0.1',
                           'dns_nameservers': ['10.0.0.2'],
                           'host_routes': [],
                           'cidr': '10.0.0.0/24',
                           'allocation_pools': allocation_pools,
                           'enable_dhcp': False,
                           'ip_version': 4}}

    def _make_floatingip_dict(self, network):
        return {"floatingip": {"floating_network_id": network["id"],
                               "tenant_id": network["tenant_id"],
                               "router_id": "testrouter"}}

    def _make_ptag_update_dict(self, name="test_tenant"):
        return {"policy_tag": {
                               "tenant_id": "test_tenant",
                               "name": name}}
